#ifndef MapEditor_h
#define MapEditor_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "MapCoordinateSystem.h"
#include "MapEditorDecoration.h"
#include "MapObject.h"
#include "MapPassInstance.h"
#include "MapLayerInstance.h"
#include "MapSelectionManager.h"
#include "ModeManager.h"
#include <Math/TDRectangle.h>
#include <Transactions/TransactionManager.h>

class MapProjection;
class wxPropertyGrid;
class MapGrid;
class MapEditor;

WX_DECLARE_LIST(MapLayerInstance,MapLayerInstanceList);
WX_DECLARE_LIST(MapEditor,MapEditorList);

WX_DECLARE_HASH_MAP( MapLayer*,
                     MapLayerInstance*,
                     wxPointerHash,
                     wxPointerEqual,
                     MapLayerInstanceMap);

WX_DECLARE_STRING_HASH_MAP(MapLayerInstance*,MapLayerNameMap);

WX_DECLARE_LIST(MapPassInstance,MapPassInstanceList);
WX_DECLARE_HASH_MAP( MapPass *,
                     MapPassInstance*,
                     wxPointerHash,
                     wxPointerEqual,
                     MapPassInstanceMap);

WX_DECLARE_STRING_HASH_MAP(MapPassInstance*,MapPassNameMap);

/**
 * \brief A MapEditor encapsulates all objects required for a map editor GUI.
 *
 * A MapEditor consists of a selection manager, a mode manager, a coordinate
 * system specification and a projection specification, manages the interaction
 * with the drawing window as well as the order and painting of map layers and
 * passes.
 *
 * \todo Design: Factor out layer and pass management.
 */
class MapEditor: public wxEvtHandler {
        
        DECLARE_CLASS(MapEditor)
        
public:
	MapEditor(wxWindow* drawingWindow,
                  MapProjection* projection,
                  TransactionManager* tmgr=new TransactionManager());
        virtual ~MapEditor();
        
        MapSelectionManager& GetSelectionManager() { return *_selectionManager; }
        ModeManager& GetModeManager() { return _modeManager; }
        MapCoordinateSystem& GetCoordinateSystem() { return _coordinateSystem; }
        TransactionManager* GetTransactionManager() { return _transactionManager; }
        
        MapProjection* GetProjection() const { return _projection; }
        
        /**
         * \brief Set new projection.
         *
         * \return old projection
         */
        MapProjection* SetProjection(MapProjection* projection);
        
        wxWindow* GetDrawingWindow() const { return _drawingWindow; }
        
        /**
         * @name Layer management
         * The list of layers is organised as a stack, where layers lower in the
         * stack are obscured by layers above them.
         *
         * When painting the map, all enabled layers are iterated in bottom-up
         * order for each enabled pass.
         */
        // @{
        
        /**
         * \brief Add a new layer instance on top of the layer stack.
         */
        MapLayerInstance* PushLayer(MapLayer* layer);
        
        /**
         * \brief Add a new layer instance below another layer.
         */
        MapLayerInstance* InsertLayer(MapLayer* layer, MapLayer* below);
        
        /**
         * \brief Find a layer instance on the layer stack based on its layer name.
         */
        MapLayerInstance* FindLayerInstanceByName(wxString name);
        
        /**
         * \brief Find a layer instance on the layer stack based on its layer.
         */
        MapLayerInstance* FindLayerInstanceByLayer(MapLayer* layer);
        
        /**
         * \brief Get the complete list of layer instances in bottom-up-order.
         */
        const MapLayerInstanceList& GetMapLayerInstances() const { return _mapLayerInstances; }
        
        /**
         * \brief Remove a layer instance from the layer stack based on its layer.
         *
         * \todo This should actually work on a layer instance, not on a layer.
         */
        void RemoveLayer(MapLayer* layer);
        // @}
        
        /**
         * @name Pass management
         * The list of passes is organised as a stack, where passes lower in the
         * stack are obscured by passes above them.
         *
         * Passes can be used to separate different features of map layers,
         * e.g. to always draw lights above pavement, or to enable and disable
         * specific parts of the map layers.
         *
         * When painting the map, all enabled passes are iterated in bottom-up
         * order, where for each pass all enabled layers are invoked to paint
         * their part for the given pass.
         */
        // @{
        /**
         * \brief Add a pass instance on top of the pass stack.
         */
        MapPassInstance* PushPass(const MapPass& pass);
        
        /**
         * \brief Add a pass instance below the given pass instance on the pass
         * stack.
         */
        MapPassInstance* InsertPass(const MapPass& pass, const MapPass& below);
        
        /**
         * \brief Find a pass instance on the pass stack by its name.
         */
        MapPassInstance* FindPassInstanceByName(wxString name);
        
        /**
         * \brief Find a pass instance on the pass stack by its pass.
         */
        MapPassInstance* FindPassInstanceByPass(const MapPass& pass);
         
        /**
         * \brief Get the complete list of pass instances in bottom-up-order.
         */
        const MapPassInstanceList& GetMapPassInstances() const { return _mapPassInstances; }
       
        /**
         * \brief Remove a pass instance from the pass stack.
         *
         * \todo This should actually work on a pass instance not on a pass.
         */
        void RemovePass(const MapPass& pass);
        // @}
        
	wxMenuItem* GetMenuItemFromCommandEvent(wxCommandEvent& evt);
        void OnToggleLayerVisibility(wxCommandEvent& evt);
        void OnToggleLayerLock(wxCommandEvent& evt);
        void OnTogglePassVisibility(wxCommandEvent& evt);
        
        /**
         * @name Canvas management
         * Canvas management is responsible for repainting parts or all
         * of the map when required by any part of the software, e.g.
         * due to a moving selection box or similar.
         */
        // @{
        /**
         * \brief Refresh the whole editor window.
         */
        void DamageAll();
        
        /**
         * \brief Refresh only the given rectangle in client coordinates.
         * 
         * The origin of the client coordinate system is the upper left corner.
         */
        void DamageRect(wxRect clientRect);
        // @}
        
        /**
         * \brief Scroll the editor view by the given device delta.
         *
         * This also notifies foreground and background decorations (e.g. grids)
         * to redraw appropriately.
         */
        void Scroll(wxSize devDelta, bool updateCanvasBounds=true);
        
        /**
         * \brief Scroll the editor view by the given delta given as fraction
         *        of the view size.
         */
        void Scroll(float deltaFractionX, float deltaFractionY);
        
        /**
         * \brief Scroll one step left.
         *
         * \sa Scroll_Step_Fraction
         */
        void ScrollLeft() { Scroll(-Scroll_Step_Fraction,0.0); }
        
        /**
         * \brief Scroll one step left.
         *
         * \sa Scroll_Step_Fraction
         */
        void ScrollRight() { Scroll(Scroll_Step_Fraction,0.0); }
        
        /**
         * \brief Scroll one step left.
         *
         * \sa Scroll_Step_Fraction
         */
        void ScrollUp() { Scroll(0.0,-Scroll_Step_Fraction); }
        
        /**
         * \brief Scroll one step left.
         *
         * \sa Scroll_Step_Fraction
         */
        void ScrollDown() { Scroll(0.0,Scroll_Step_Fraction); }
        
        /**
         * \brief Zoom the view by the given factor, centered around the given
         *        center.
         *
         * \param localCenter   The fixpoint/center point in local coordinates
         * \param factor        The zoom factor (>1: zoom in, <1: zoom out)
         */
        void Zoom(Point3D localCenter, double factor);
        
        /**
         * \brief Zoom the editor view by a factor, centered around the view
         *        center.
         *
         * \param factor        The zoom factor (>1: zoom in, <1: zoom out)
         */
        void Zoom(double factor);
        
        /**
         * \brief Zoom in by a given number of steps, centered around the given
         *        center.
         *
         * \param localCenter   The fixpoint/center point in local coordinates
         * \param steps         Number of steps to zoom by (negative=>zoom out,
         *                                                  positive=>zoom in)
         *
         * \sa Zoom_Step_Factor
         */
        void Zoom(Point3D localCenter, int steps) {
                Zoom(localCenter,pow(Zoom_Step_Factor,steps));
        }
        /**
         * \brief Zoom in by a given number of steps.
         *
         * \param steps         Number of steps to zoom by (negative=>zoom out,
         *                                                  positive=>zoom in)
         *
         * \sa Zoom_Step_Factor
         */
        void Zoom(int steps) {
                Zoom(pow(Zoom_Step_Factor,steps));
        }
        
        /**
         * \brief Set the zoom factor.
         *
         * \param zoom          The zoom factor (=1: local units=device units)
         */
        void SetZoom(double zoom);
        
        /**
         * \brief Zoom and pan the view so that a given region is contained in
         * the view.
         *
         * \param targetBounds  The bounds in local coordinates.
         */
        void ZoomTo(TDRectangle targetBounds);
        
        /**
         * \brief Zoom and pan the view so that a given region is contained in
         * the view.
         *
         * \param targetBounds  The bounds in geographical coordinates.
         */
        void ZoomToGeo(TDRectangle targetBounds);
        
        void AdjustScrollbars();
        
        void OnScrollWin(wxScrollWinEvent& event);
        
        /**
         * \brief Get the bounds of all objects displayed in the editor in local
         * coordinates.
         */
        TDRectangle GetLocalBounds();
        
        /**
         * \brief Get the bounds of all objects displayed in the editor in
         * geographic coordinates.
         */
        TDRectangle GetGeographicBounds();
        
        /**
         * \brief Get the local bounds of the currently visible region.
         */
        TDRectangle GetVisibleLocalBounds();
        
        /**
         * \brief Get the geographic bounds of the currently visible region.
         */
        TDRectangle GetVisibleGeographicBounds();
        
        /**
         * \brief Get a list of objects of the given class from this layer.
         *
         * \todo Design: This is actually bad design as this really only
         *       provided by MapObjectContainer.
         */
         void GetObjectsOfClass(MapObjectList& list, wxClassInfo* targetClass);
        
        /**
         * \brief Get the list of all visible objects intersecting a given
         * rectangle.
         *
         * \param list          List to be filled with intersecting objects.
         * \param r             Rectangle to be intersected with visible objects.
         */
        void GetHitObjects(MapObjectList& list,
                           TDRectangle r,
                           bool includeIntersected=true);
        
        /* Provide actions for the action menu */
        void ProvideActionMenu(wxMenu& menu);
        
        /* Paint the editor window */
        void OnPaint(wxPaintEvent& evt);
        
        /* Handle resizing of the editor window */
        void OnResize(wxSizeEvent& evt);
        
        void OnMouseEnter(wxMouseEvent& evt);
        
        /**
         * \brief Return the MapEditorDecoration object representing the map grid.
         */
        MapGrid* GetGrid() const {
                return _grid;
        }
        
        /**
         * \brief Return a list of all currently open editors.
         */
        static const MapEditorList& GetEditorList() {
                return _editorList;
        }
        
        /**
         * \brief Fraction of the view size to scroll per step.
         *
         * \sa ScrollUp(), ScrollDown(), ScrollLeft(), ScrollRight()
         */
        static float Scroll_Step_Fraction;
        
        /**
         * \brief Factor by which to zoom in per step.
         *
         * The inverse is the factor by which to zoom out per step.
         *
         * \sa ZoomIn(), ZoomOut()
         */
        static float Zoom_Step_Factor;
        
        static wxColour Map_Background_Colour;
        
        static float Scroll_Line_Width,Scroll_Line_Height;
protected:
        TransactionManager* _transactionManager;
        MapEditorDecorationList _foregroundDecos,_backgroundDecos;
        MapSelectionManager::SmartRef _selectionManager;
        ModeManager _modeManager;
        MapCoordinateSystem _coordinateSystem;

        MapLayerInstanceList _mapLayerInstances;
        MapLayerInstanceMap _mapLayerInstanceMap;
        MapLayerNameMap _mapLayerNameMap;

        MapPassInstanceList _mapPassInstances;
        MapPassInstanceMap _mapPassInstanceMap;
        MapPassNameMap _mapPassNameMap;
        
        MapProjection* _projection;
        wxWindow* _drawingWindow;
        
        MapGrid* _grid;
        
        /**
         * The canvas bounds are the bounds of the canvas as used to define
         * the range of the scrollbars.
         */
        TDRectangle _canvasLocalBounds;
        
        static MapEditorList _editorList;
        
        DECLARE_EVENT_TABLE();
};

#endif
