// MapProjection.h - An abstract base class to encapsulate a projection.
//
// Written by David Luff, started 2003.
//
// Copyright (C) 2002  David C Luff - david.luff@nottingham.ac.uk
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef MapProjection_h
#define MapProjection_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include <Math/point3d.h>

class MapProjectionDescriptor;

/**
 * \brief A MapProjection allows to convert geographic coordinates to logical
 * coordinates according to a given projection.
 *
 * The geographic coordinates are assumed to follow the conventions of WGS84.
 *
 * See \ref coordinate_systems for a more detailed description of the coordinate
 * systems.
 */
class MapProjection: public wxObject {
        DECLARE_ABSTRACT_CLASS(MapProjection);
        
        friend class MapProjectionDescriptor;
public:
        virtual ~MapProjection();
        
        /**
         * \brief Setup the projection with the given centre.
         *
         * The projection can use the centre information to select parameters for
         * the projection which are suited for the area directly around centre
         * (e.g., the UTM zone).
         *
         * \return true if initialisiation is successful
         */
        virtual bool Init(Point3D centre)=0;
        
        /**
         * \brief Convert geographic coordinates to logical coordinates according
         * to this projection.
         *
         * \param pt        A point in geographic coordinates.
         *
         * \return the projected point in logical coordinates.
         */
        virtual Point3D ConvertToLocal(Point3D pt)=0;
        
        /**
         * \brief Convert logical coordinates to geographic coordinates according
         * to this projection.
         *
         * \param pt        A point in logical coordinates.
         *
         * \return the unprojected point in geographic coordinates.
         */
        virtual Point3D ConvertFromLocal(Point3D pt)=0;
        
        /**
         * \brief Get the MapProjectionDescriptor this projection was created
         * from.
         */
        MapProjectionDescriptor* GetDescriptor() const { return _descriptor; }
protected:
        MapProjectionDescriptor* _descriptor;
        
        void SetDescriptor(MapProjectionDescriptor* descriptor) {
                _descriptor=descriptor;
        }
};

#endif	// MapProjection_h
