#ifndef Mode_h
#define Mode_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "ModeManager.h"

#include "ModeChangeListener.h"

WX_DECLARE_LIST(ModeChangeListener,ModeChangeListenerList);

/**
 * \brief A Mode represents a specific operation mode of a MapEditor.
 *
 * It is passed events from the drawing window via the ModeManager and can
 * handle them or pass them down in the mode stack.
 *
 * Event handling is defined as for any other wxWidgets event handler using the
 * wx macros.
 */
class Mode : public wxEvtHandler {
        
        DECLARE_ABSTRACT_CLASS(Mode)
        
        friend class ModeManager;
        
public:
	Mode();
	virtual ~Mode();
        
        /**
         * \brief Set the mode manager for this mode.
         */
        void SetModeManager(ModeManager* manager) { _manager=manager; }
        
        /**
         * \brief Get the mode manager for this mode.
         */
	ModeManager* GetModeManager() const { return _manager; }
        
        /**
         * \brief Convenience method to get the map editor for this mode.
         */
	MapEditor* GetEditor() const { return (_manager?_manager->GetEditor():NULL); }
	
        /**
         * \brief Enter this mode.
         *
         * The mode can veto being entered by returning false. This could, e.g.,
         * be useful when the mode wants to propose a replacement mode for itself
         * under certain circumstances, using ModeManager::InsertAbove() or similar.
         *
         * \return false, if the mode wants to veto being entered.
         */
        virtual bool Enter();
        
        /**
         * \brief Determine whether this mode can be exited.
         *
         * By returning false the mode can veto removal from the stack, e.g.,
         * if an operation is neither finished nor cancelled.
         *
         * \return true if and only if this mode can be exited.
         */
	virtual bool CanExit();
        
        /**
         * \brief Leave this mode.
         *
         * \param force         Flag indicating whether the mode shall be
         *                      forcibly left, e.g., when the mode is cancelled.
         */
	virtual void Leave(bool force=false);
        
        /**
         * \brief Get the status bar string for this mode.
         */
	virtual wxString GetStatus();
	
        /**
         * \brief Convenience method to trigger updating of the status bar.
         */
	void UpdateStatus();

        /**
         * \brief Convenience method to non-forcibly end this mode
         */
	void Done();
        
        void RegisterModeChangeListener(ModeChangeListener* listener) {
                _changeListeners.Append(listener);
        }
        
        void UnregisterModeChangeListener(ModeChangeListener* listener) {
                _changeListeners.DeleteObject(listener);
        }
protected:
        void FireBeforeEnteringMode() {
                ModeChangeListenerList::Node *listNode;
                
                for (listNode=_changeListeners.GetFirst();listNode;listNode=listNode->GetNext()) {
                        listNode->GetData()->BeforeEnteringMode(this);
                }
        }

        void FireAfterLeavingMode() {
                ModeChangeListenerList::Node *listNode;
                
                for (listNode=_changeListeners.GetFirst();listNode;listNode=listNode->GetNext()) {
                        listNode->GetData()->AfterLeavingMode(this);
                }
        }
private:
        DECLARE_EVENT_TABLE()
        
        ModeChangeListenerList _changeListeners;
        
        ModeManager* _manager;
};

#endif
