#ifndef TDRectangle_h
#define TDRectangle_h
#include <algorithm>
#include "point3d.h"
#include <float.h>
#include <iostream>

/*
 * Class for representing rectangles.
 *
 * The nomenclature assumes positive y upwards/norther and
 * positive x left/eastern.
 */
class TDRectangle {
public:
        TDRectangle():_left(DBL_MAX),_right(-DBL_MAX),_top(-DBL_MAX),_bottom(DBL_MAX) {}
        
        TDRectangle(const TDRectangle& r):
                _left(r._left),_right(r._right),_top(r._top),_bottom(r._bottom) {}
        
        TDRectangle(double top, double bottom, double right, double left):
                _left(left),_right(right),_top(top),_bottom(bottom) {}
        
        TDRectangle(Point3D p):
                _left(p.x()),
                _right(p.x()),
                _top(p.y()),
                _bottom(p.y()) {}
        
        TDRectangle(Point3D p1, Point3D p2):
                _left(std::min(p1.x(),p2.x())),
                _right(std::max(p1.x(),p2.x())),
                _top(std::max(p1.y(),p2.y())),
                _bottom(std::min(p1.y(),p2.y())) {}
        
        TDRectangle(Point3D center, double w, double h):
                _left(center.x()-w/2),
                _right(center.x()+w/2),
                _top(center.y()+h/2),
                _bottom(center.y()-h/2) {}
        
        double GetTop()    const { return _top; }
        double GetBottom() const { return _bottom; }
        double GetLeft()   const { return _left; }
        double GetRight()  const { return _right; }

        double GetNorth()  const { return _top; }
        double GetEast()   const { return _right; }
        double GetSouth()  const { return _bottom; }
        double GetWest()   const { return _left; }
        
        void SetTop(double v)    { _top=v; }
        void SetBottom(double v) { _bottom=v; }
        void SetLeft(double v)   { _left=v; }
        void SetRight(double v)  { _right=v; }
        
        void SetNorth(double v) { _top=v; }
        void SetEast(double v)  { _right=v; }
        void SetSouth(double v) { _bottom=v; }
        void SetWest(double v)  { _left=v; }
        
        Point3D GetTopLeft() const { return Point3D(_left,_top,0); }
        Point3D GetTopRight() const { return Point3D(_right,_top,0); }
        Point3D GetBottomLeft() const { return Point3D(_left,_bottom,0); }
        Point3D GetBottomRight() const { return Point3D(_right,_bottom,0); }
        Point3D GetSize() const { return Point3D(_right-_left,_top-_bottom,0); }
        
        Point3D GetCenter() const {
                return Point3D((_left+_right)/2,(_top+_bottom)/2,0);
        }
        
        double GetWidth() const { return _right-_left; }
        
        void SetWidth(double w) {
                double cx=(_left+_right)/2.0;
                _left=cx-w/2.0;
                _right=cx+w/2.0;
        }
        
        double GetHeight() const { return _top-_bottom; }
        
        void SetHeight(double h) {
                double cy=(_top+_bottom)/2.0;
                _top=cy+h/2.0;
                _bottom=cy-h/2.0;
        }
        
        void Enlarge(double dx, double dy) {
                _left-=dx/2;
                _right+=dx/2;
                _top+=dy/2;
                _bottom-=dy/2;
        }
        
        void Translate(double dx, double dy) {
                _top+=dy;
                _bottom+=dy;
                _left+=dx;
                _right+=dx;
        }
        
        void Scale(double sx, double sy) {
                double cx=(_left+_right)/2.0;
                double cy=(_top+_bottom)/2.0;
                
                _top=(_top-cy)*sy+cy;
                _bottom=(_bottom-cy)*sy+cy;
                _left=(_left-cx)*sx+cx;
                _right=(_right-cx)*sx+cx;
        }
        
        bool IsEmpty() const {
                return (_top<_bottom) || (_right<_left);
        }
        
        bool Intersects(const TDRectangle r) const {
                return !TDRectangle::Intersection(*this,r).IsEmpty();
        }
        
        bool Contains(const Point3D p) const {
                return (_left<=p.x() && _right>=p.x()) &&
                                (_bottom<=p.y() && _top>=p.y());
        }
        
        bool Contains(const TDRectangle r) const {
                return (r._top<=_top) && (r._bottom>=_bottom) &&
                        (r._right<=_right) && (r._left>=_left);
        }
        
        static TDRectangle Intersection(const TDRectangle a, const TDRectangle b) {
                return TDRectangle(std::min(a._top,b._top),
                                 std::max(a._bottom,b._bottom),
                                 std::min(a._right,b._right),
                                 std::max(a._left,b._left));
        }
        
        static TDRectangle Union(const TDRectangle a, const TDRectangle b) {
                return TDRectangle(std::max(a._top,b._top),
                                 std::min(a._bottom,b._bottom),
                                 std::max(a._right,b._right),
                                 std::min(a._left,b._left));
        }
protected:
        double _left,_right,_top,_bottom;
};
#endif // TDRectangle_h
