#ifndef Transaction_h
#define Transaction_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "TransactionObject.h"

WX_DECLARE_HASH_MAP( TransactionObject*,
                     wxObject*,
                     wxPointerHash,
                     wxPointerEqual,
                     MementoMap);

/**
 * \brief A Transaction encapsulates a set of changes which can be collectively
 * undone or redone.
 *
 * A transaction stores a list of memento objects used to restore the previous
 * state of all objects modified during the transaction.
 *
 * See \ref transaction_concept for more information.
 *
 * \sa TransactionManager TransactionObject
 */
class Transaction: public wxObject {
        
        DECLARE_CLASS(Transaction)
        
        friend class TransactionManager;
        
public:
        bool IsCommitted() const { return _committed; }
        
        const wxString& GetDescription() const { return _description; }
protected:
	Transaction();
        virtual ~Transaction();
        
        /**
         * \brief Store a copy of the given object in this transaction.
         *
         * This is typically called by a TransactionObject when it is about
         * to be modified for the first time in a transaction.
         */
        void StoreCopy(TransactionObject* obj);
        
        /**
         * \brief Delete the copy of the given object from this transaction, if any.
         *
         * This is called by a TransactionObject when it is about to be destroyed.
         *
         * \todo Destroy the transaction when it contains no further mementos.
         */
        void DeleteCopy(TransactionObject* obj);
        
        /**
         * \brief Commit this transaction.
         *
         * \param description   Text to be used to describe this transaction
         *                      (e.g. in the undo and redo menu)
         */
        void Commit(const wxString& description=wxT(""));
        
        /**
         * \brief Restore all objects to the state at the beginning of this
         * transaction
         */
        void Restore();
        
        /**
         * \brief Backup the current state of all objects in this transaction
         *        to another transaction.
         *
         * During undo of the top transaction on the stack, this method is
         * used to store the current state of all restored objects in order
         * to be able to redo the changes undone.
         *
         * \param target        The transaction to save to.
         */
        void AddCurrentCopiesToTransaction(Transaction* target);

        bool _committed;
        wxString _description;
        
        MementoMap _mementos;
};

#endif

