#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "TransactionManager.h"
#include "Transaction.h"

#include <wx/listimpl.cpp>

IMPLEMENT_CLASS(TransactionManager,wxObject);

WX_DEFINE_LIST(TransactionStackChangeListenerList);

TransactionManager::TransactionManager(TransactionStackSize backlogSize):
        _stackPtr(0),
        _maxBacklogSize(backlogSize),
        _holdCount(0),
	_editorCount(0)
{
}

TransactionManager::~TransactionManager() {
        DestroyAllFrom(0);
}

void TransactionManager::Clear() {
	Commit();	// make sure the current transaction is properly committed
        _stackPtr=0;
        DestroyAllFrom(0);
}

void TransactionManager::StoreCopy(TransactionObject* obj) {
        if (_stackPtr==0 || _stack[_stackPtr-1]->IsCommitted())
                OpenNewTransaction();
        _stack[_stackPtr-1]->StoreCopy(obj);
}

void TransactionManager::DeleteCopy(TransactionObject* obj) {
        TransactionStack::iterator it;
        for (it=_stack.begin();it!=_stack.end();it++) {
                (*it)->DeleteCopy(obj);
	}
}

void TransactionManager::Commit(const wxString& description) {
        if (_stackPtr==0 || _stack[_stackPtr-1]->IsCommitted())
                return;
        _stack[_stackPtr-1]->Commit(description);
}

void TransactionManager::Abort() {
        if (_stackPtr==0 || _stack[_stackPtr-1]->IsCommitted())
                return;
        /*
         * We assume that we're on top of the stack, as this is an uncommitted
         * transaction which can only originate from OpenNewTransaction(),
         * which in turn clears the stack above the stack pointer before
         * creating the new transaction.
         */
        wxASSERT(_stackPtr==_stack.size());
        _stack[_stackPtr-1]->Restore();
        delete _stack[_stackPtr-1];
        _stack.pop_back();
        _stackPtr--;
        FireStackChange();
}

void TransactionManager::Undo() {
        if (!CanUndo())
                return;
        /*
         * Make sure the current transaction is committed, so we
         * can distinguish the backup transaction, which is not committed.
         */
        Commit();
        /*
         * We need to add a backup of the values to be overwritten
         * in the transaction above the current to be able to redo
         * the transaction fully.
         * If we are at the top transaction, we need to create
         * a backup transaction.
         */
        if (_stackPtr==_stack.size()) {
                Transaction* backup;
                backup=new Transaction();
                _stack.push_back(backup);
        }
        wxASSERT(_stackPtr<_stack.size());
        /*
         * Store copies of the objects to be restored in the
         * transaction above the current, if no copies do yet exist.
         */
        _stack[_stackPtr-1]->AddCurrentCopiesToTransaction(_stack[_stackPtr]);
        _stack[_stackPtr-1]->Restore();
        _stackPtr--;
        FireStackChange();
}

void TransactionManager::Redo() {
        if (!CanRedo())
                return;
        /*
         * See the general comments on this class in the documentation
         * on why we're actually working on index _stackPtr+1.
         */
        wxASSERT(_stackPtr+1<_stack.size());
        _stack[_stackPtr+1]->Restore();
        if (_stack[_stackPtr]->IsCommitted()) {
                _stackPtr++;
                FireStackChange();
        }
}

const wxString& TransactionManager::GetUndoDescription(size_t idx) const {
        wxASSERT(0<=idx && idx<GetUndoSteps());
        return _stack[_stackPtr-1-idx]->GetDescription();
}

const wxString& TransactionManager::GetRedoDescription(size_t idx) const {
        wxASSERT(0<=idx && idx<GetRedoSteps());
        return _stack[_stackPtr+idx]->GetDescription();
}

void TransactionManager::OpenNewTransaction() {
        Transaction* trans;
        
        /* Go back if this is the backup transaction */
        /* FIXME: Can this actually happen? */
        if (_stackPtr==_stack.size() && _stackPtr && !_stack[_stackPtr-1]->IsCommitted())
                _stackPtr--;
        /* Drop all transactions after the current */
        DestroyAllFrom(_stackPtr);
        
        /* Shave the transaction stack to the maximum size */
        if (_stack.size()==_maxBacklogSize) {
                Transaction* trans=_stack[0];
                _stack.pop_front();
                delete trans;
        }
        
        trans=new Transaction();
        _stack.push_back(trans);
        _stackPtr=_stack.size();
        FireStackChange();
}

void TransactionManager::DestroyAllFrom(TransactionStackPtr start) {
        /* If we destroy before the current stack pointer,
         * we must destroy all transactions before the stack pointer.
         * Otherwise we would break the chain of incremental
         * modifications.
         */
        wxASSERT(start==0 || _stackPtr<=start);
        while (_stack.size()>start) {
                Transaction* trans=_stack.back();
                _stack.pop_back();
                delete trans;
        }
        if (start==0)
                _stackPtr=0;
        FireStackChange();
}

void TransactionManager::FireStackChange() {
	TransactionStackChangeListenerList::Node *node;
	for (node=_transactionChangeListeners.GetFirst();node;node=node->GetNext()) {
		node->GetData()->OnTransactionStackChange(this);
	}
}
