#ifndef TransactionObject_h
#define TransactionObject_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "TransactionManager.h"
#include "TransactionObjectChangeListener.h"

WX_DECLARE_LIST(TransactionObjectChangeListener,TransactionObjectChangeListenerList);

/**
 * \brief A TransactionObject is an object whose state can be stored in a
 * transaction.
 *
 * Each transaction object is associated with a TransactionManager and stores
 * a dirty flag and a reference count.
 *
 * The reference count is used to determined whether a transaction object is
 * still in use, e.g., after destroying a transaction during backlog consolidation.
 *
 * Note that a memento of an object being stored in a live transaction also
 * implies a reference to the object, i.e., an object is not destroyed as long
 * as it is part of the state of a live transaction.
 *
 * The dirty flag is used to determine whether the object was already modified
 * during the current transaction.
 *
 * See \ref transaction_concept for more information.
 *
 * \sa TransactionManager Transaction
 *
 * \todo Introduce automatic activation and deactivation of transaction objects.
 */
class TransactionObject: public wxObject {
        
        DECLARE_ABSTRACT_CLASS(TransactionObject)
        
        friend class Transaction;
        
public:
	TransactionObject(TransactionManager* manager);
        virtual ~TransactionObject();
        
        /**
         * \brief Create a memento for this object.
         *
         * This is typically called from Transation::StoreCopy() to
         * retrieve a memento for this object.
         *
         * The memento should store all information necessary to restore
         * the persistent part of this object to the current state.
         *
         * \return the memento
         */
         virtual wxObject* CreateMemento()=0;
        
        /**
         * \brief Restore the state of this object from a memento.
         *
         * This is typically called from Transaction::Restore() to restore
         * this object to a previous state.
         *
         * The memento was previously created by CreateMemento().
         */
         virtual void RestoreMemento(wxObject*) {};
        
        /**
         * \brief Determine whether this object is dirty.
         *
         * An object being dirty means that it was already modified during
         * the current transaction and a copy of its state at the beginning
         * of the transaction is already stored.
         */
        bool IsDirty() const { return _dirty; }
        
        /**
         * \brief Register an object change listener on this object.
         *
         * The MapObjectChangeListener is triggered before and after an attribute of
         * this object changes.
         *
         * \sa UnregisterChangeListener() BeforeWrite() AfterWrite()
         */
        void RegisterChangeListener(TransactionObjectChangeListener* listener) {
                 _changeListeners.Append(listener);
        }
        
        /**
         * \brief Unregister an object change listener from this object.
         *
         * \sa RegisterChangeListener()
         */
        void UnregisterChangeListener(TransactionObjectChangeListener* listener) {
                 _changeListeners.DeleteObject(listener);
        }
        
        /**
         * \brief Increase the reference count of this object.
         *
         * \todo Document SmartRefs
         */
        void Ref() { _refCount++; }
        
        /**
         * \brief Decrease the reference count of this object.
         */
        void UnRef() {
                wxASSERT_MSG(_refCount>0,
                        wxT("Trying to deref non-referenced transaction object"));
                if (--_refCount==0)
                        delete this;
        }
        
        unsigned int GetRefCount() const { return _refCount; }
        
        TransactionManager* GetTransactionManager() const { return _manager; }
protected:
        /**
         * \brief Trigger TransactionObjectChangeListener::BeforeObjectChange()
         *        on all registered object change listeners.
         */
        void TriggerBeforeObjectChange();
        
        /**
         * \brief Trigger TransactionObjectChangeListener::AfterObjectChange()
         *        on all registered object change listeners.
         */
        void TriggerAfterObjectChange();
        
        /**
         * \brief Trigger all necessary actions before a change to the object.
         *
         * These actions are:
         *
         * -# Trigger TransactionObjectChangeListener::BeforeObjectChange() on all
         *    registered object change listeners.
         * -# Ensure storage of a memento
         */
        virtual void BeforeWrite();
        
        /**
         * \brief Trigger all necessary actions after a change to the object.
         *
         * These actions are:
         *
         * -# Trigger MapObjectChangeListener::AfterObjectChange() on all
         *    registered object change listeners.
         */
        virtual void AfterWrite();

        /**
         * \brief Convenience class to ease uniform definition of Mementos.
         */
        class Memento: public wxObject {
                friend class TransactionObject;
        public:
                Memento(const TransactionObject&) {}
        };
private:
        unsigned int _refCount;
        TransactionManager* _manager;
        bool _dirty;
        TransactionObjectChangeListenerList _changeListeners;
};

#define TRANSACTIONOBJECT_STATE _myState

#define TRANSACTIONOBJECT_DEFINE_STATE(STATETYPE) \
        protected: \
                STATETYPE TRANSACTIONOBJECT_STATE;

#define TRANSACTIONOBJECT_DEFINE_STATE_MEMENTO(OBJECT,SUPERCLASS,STATETYPE) \
        protected: \
                TRANSACTIONOBJECT_DEFINE_STATE(STATETYPE) \
                \
                class Memento: public SUPERCLASS::Memento { \
                        friend class OBJECT; \
                public: \
                        Memento(const OBJECT& obj): \
                                SUPERCLASS::Memento(obj) \
                        { \
                                _savedState=obj.TRANSACTIONOBJECT_STATE; \
                        } \
                protected: \
                        STATETYPE _savedState; \
                }; \
        public: \
                virtual Memento* CreateMemento() { \
                        return new Memento(*this); \
                } \
                virtual void RestoreMemento(wxObject* object) { \
                        SUPERCLASS::RestoreMemento(object); \
                        Memento* memento=dynamic_cast<Memento*>(object); \
                        wxASSERT(memento); \
                        \
                        TRANSACTIONOBJECT_STATE=memento->_savedState; \
                }

#define TRANSACTIONOBJECT_STATEVAR(STATEVAR) TRANSACTIONOBJECT_STATE.STATEVAR

#define TRANSACTIONOBJECT_DEFINE_STATE_GETTER(TYPE,SUFFIX,STATEVAR) \
public: \
        TYPE Get##SUFFIX () const { \
                return TRANSACTIONOBJECT_STATEVAR(STATEVAR); \
        } \

#define TRANSACTIONOBJECT_DEFINE_STATE_SETTER(TYPE,SUFFIX,STATEVAR) \
public: \
        void Set##SUFFIX (TYPE v) { \
                BeforeWrite(); \
                 \
                TRANSACTIONOBJECT_STATEVAR(STATEVAR)=v; \
                 \
                AfterWrite(); \
        }

#define TRANSACTIONOBJECT_DEFINE_STATE_SETTER_WCOMP(TYPE,SUFFIX,STATEVAR) \
        TRANSACTIONOBJECT_DEFINE_STATE_SETTER_WUCOMP(TYPE,SUFFIX,STATEVAR,TRANSACTIONOBJECT_STATEVAR(STATEVAR)==v)

#define TRANSACTIONOBJECT_DEFINE_STATE_SETTER_WUCOMP(TYPE,SUFFIX,STATEVAR,UCOMP) \
public: \
        void Set##SUFFIX (TYPE v) { \
                if (UCOMP) \
                        return; \
                BeforeWrite(); \
                 \
                TRANSACTIONOBJECT_STATEVAR(STATEVAR)=v; \
                 \
                AfterWrite(); \
        }

#define TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR(TYPE,SUFFIX,STATEVAR) \
        TRANSACTIONOBJECT_DEFINE_STATE_GETTER(TYPE,SUFFIX,STATEVAR) \
        TRANSACTIONOBJECT_DEFINE_STATE_SETTER(TYPE,SUFFIX,STATEVAR)

#define TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(TYPE,SUFFIX,STATEVAR) \
        TRANSACTIONOBJECT_DEFINE_STATE_GETTER(TYPE,SUFFIX,STATEVAR) \
        TRANSACTIONOBJECT_DEFINE_STATE_SETTER_WCOMP(TYPE,SUFFIX,STATEVAR)

#define TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WUCOMP(TYPE,SUFFIX,STATEVAR,UCOMP) \
        TRANSACTIONOBJECT_DEFINE_STATE_GETTER(TYPE,SUFFIX,STATEVAR) \
        TRANSACTIONOBJECT_DEFINE_STATE_SETTER_WUCOMP(TYPE,SUFFIX,STATEVAR,UCOMP)

#define TRANSACTIONOBJECT_DEFINE_SMARTREF(OBJECT) \
public: \
        class SmartRef { \
        public: \
                SmartRef(OBJECT* obj):_object(obj) { \
                        if (_object) {\
                                _object->TransactionObject::Ref(); \
                        } \
                } \
                SmartRef(const SmartRef& other):_object(other._object) { \
                        if (_object) {\
                                _object->TransactionObject::Ref(); \
                        } \
                } \
                SmartRef():_object(NULL) {} \
                ~SmartRef() { \
                        if (_object) {\
                                _object->TransactionObject::UnRef(); \
                        } \
                } \
                SmartRef& operator=(OBJECT* obj) { \
                        if (_object) {\
                                _object->TransactionObject::UnRef(); \
                        } \
                        _object=obj; \
                        if (_object) {\
                                _object->TransactionObject::Ref(); \
                        } \
                        return *this; \
                } \
                SmartRef& operator=(const SmartRef& other) { \
                        return (*this=other._object); \
                } \
                OBJECT& operator*() const { \
                        return *_object; \
                } \
                OBJECT* operator->() const { \
                        return _object; \
                } \
                operator OBJECT*() const { \
                        return _object; \
                } \
        private: \
                OBJECT* _object; \
        };
#endif
// :mode=c++:

